import sys
import os
import time
import copy
import tempfile
from .comms import ZMQLogSink, ZMQClient
from .proc import ServerProcess
import peerprint.server.proto.state_pb2 as spb
import peerprint.server.proto.jobs_pb2 as jpb
import peerprint.server.proto.peers_pb2 as ppb
from google.protobuf.any_pb2 import Any
from enum import Enum

class ChangeType(Enum):
    JOBS = 0
    PEERS = 1

class PeerPrintQueue():
    def __init__(self, opts, peer_id, codec, binary_path, update_cb, logger, keydir=None):
        self._logger = logger
        self._opts = opts
        self._peer_id = peer_id
        self._binary_path = binary_path
        self._proc = None
        self._ready = False
        self._codec = codec
        self._update_cb = update_cb

        self._zmqlogger = None
        self._zmqclient = None
        self._proc = None

        # These are cached from updates
        self._jobs = dict()
        self._peers = ppb.PeersSummary(peer_estimate=0, variance=float('inf'), sample=[])
        # Using a temporary directory allows running multiple instances/queues
        # using the same filesystem (e.g. for development or containerized
        # farms)
        self.tmpdir = tempfile.TemporaryDirectory()
        if self._opts.zmqlog is None:
            self._opts.zmqlog = f"ipc://{self.tmpdir.name}/log.ipc"
        if self._opts.zmq is None:
            self._opts.zmq = f"ipc://{self.tmpdir.name}/cmd.ipc"
        if self._opts.zmqpush is None:
            self._opts.zmqpush = f"ipc://{self.tmpdir.name}/push.ipc"
        
        if keydir is not None:
            self._opts.privkeyfile = os.path.join(keydir, f"{opts.queue}_priv.key")
            self._opts.pubkeyfile = os.path.join(keydir, f"{opts.queue}_pub.key")

    def connect(self):
        self._zmqlogger = ZMQLogSink(self._opts.zmqlog, self._logger)
        self._proc = ServerProcess(self._opts, self._binary_path, self._logger)
        self._zmqclient = ZMQClient(self._opts.zmq, self._opts.zmqpush, self._update, self._logger)

    def destroy(self):
        if self._zmqlogger is not None:
            self._zmqlogger.destroy()
        if self._zmqclient is not None:
            self._zmqclient.destroy()
        if self._proc is not None:
            self._proc.destroy()
        self.tmpdir.cleanup()


    def _update(self, msg):
        expiry_ts = time.time() - 1*60*60 # TODO encode somewhere. manifest?
        if isinstance(msg, spb.State):
            newjobs = dict()
            for k,v in msg.jobs.items():
                newjobs[k] = self._codec.decode(v.data, v.protocol)
                newjobs[k]['peer_'] = v.owner
                if v.lock is not None and v.lock.created > expiry_ts:
                    newjobs[k]['acquired'] = True
                    newjobs[k]['acquired_by_'] = v.lock.peer
            # Set _jobs BEFORE calling update_cb in case it invokes get_jobs()
            oldjobs = copy.deepcopy(self._jobs)
            self._jobs = newjobs
            print("New job data:", self._jobs)
            if self._update_cb is not None:
                self._update_cb(ChangeType.JOBS, self._jobs, newjobs)
        elif isinstance(msg, ppb.PeersSummary):
            # Set _peers BEFORE calling update_cb in case it invokes get_peers()
            oldpeers = copy.deepcopy(self._peers)
            self._peers = msg
            if self._update_cb is not None:
                self._update_cb(ChangeType.PEERS, self._peers, msg)
        self._ready = True

    def is_ready(self):
        return self._ready

    def get_id(self):
        return self._peer_id

    # ==== Mutation methods ====

    def syncPeer(self, state: dict, addr=None):
        self._zmqclient.call(ppb.PeerStatus()) #state=state, addr=addr))
    
    def getPeers(self):
        return self._peers

    def getPeer(self, peer):
        raise NotImplementedError()
        # return self._zmqclient.call(ppb.GetPeersRequest(peerFilter=peer))

    def hasJob(self, jid) -> bool:
        return self._jobs.get(jid) is not None

    def setJob(self, jid, manifest: dict, addr=None):
        data, protocol = self._codec.encode(manifest)
        self._zmqclient.call(jpb.SetJobRequest(
            job=jpb.Job(
                id=jid,
                owner=addr,
                protocol=protocol,
                data=data,
            ),
        ))

    def getJobs(self):
        return self._jobs

    def removeJob(self, jid: str):
        self._zmqclient.call(jpb.DeleteJobRequest(id=jid))

    def acquireJob(self, jid: str):
        self._zmqclient.call(jpb.AcquireJobRequest(id=jid))

    def releaseJob(self, jid: str):
        self._zmqclient.call(jpb.ReleaseJobRequest(id=jid))




import unittest
import logging
import signal
from unittest.mock import MagicMock, call, patch, ANY
from .proc import ServerProcessOpts, DependentProcess, IPFSDaemonProcess, ServerProcess

logging.basicConfig()

class TestOpts(unittest.TestCase):
    def testRender(self):
        opts = ServerProcessOpts(addr="foo", local=True)
        self.assertEqual(opts.render("bar"), ["bar", "-addr=foo", "-local=True"])

class TestDependentProcess(unittest.TestCase):
    def testDestroy(self):
        dp = DependentProcess()
        dp._proc = MagicMock(returncode=None)
        dp._logger = logging.getLogger()
        dp.destroy()
        dp._proc.send_signal.assert_has_calls([
            call(signal.SIGINT),
            call(signal.SIGKILL),
            call(signal.SIGTERM),
            ])

    def testDestroyNoProcess(self):
        dp = DependentProcess()
        dp.destroy()

class TestIPFSDaemonProcess(unittest.TestCase):
    def testInit(self):
        with patch('peerprint.wan.proc.subprocess') as subp:
            p = MagicMock(returncode=1) # Return error when calling `ipfs stats bw` to check server state, indicates not running
            subp.Popen.return_value = p
            d = IPFSDaemonProcess(logging.getLogger())
            self.assertEqual(subp.Popen.call_args[0][0][:2], ["ipfs", "daemon"])
    
    def testInitAlreadyRunning(self):
        with patch('peerprint.wan.proc.subprocess') as subp:
            p = MagicMock(returncode=0) # Return success when calling `ipfs stats bw`, indicates server is running
            subp.Popen.return_value = p
            d = IPFSDaemonProcess(logging.getLogger())
            subp.Popen.assert_called_once_with(["ipfs", "stats", "bw"], stdout=ANY, stderr=ANY)

class TestServerProcses(unittest.TestCase):
    def testInit(self):
        with patch('peerprint.wan.proc.subprocess') as subp:
            subp.Popen.return_value = None
            s = ServerProcess(ServerProcessOpts(addr="foo", local=True), "bin", logging.getLogger())
            subp.Popen.assert_called_once() # options rendering tested earlier in this file



import logging
import cmd
import traceback
import datetime
import json

class Shell(cmd.Cmd):
    intro = 'Type help to list commands. Ctrl+C to exit\n'
    prompt = '>> '

    RESULT_TYPES = ("success", "failure", "cancelled")

    class OutputCapture:
      def __init__(self):
        self.out = ""
      def write(self, s):
        self.out += s
      def dump(self):
        s = self.out
        self.out = ""
        return s

    def attach(self, lan):
      self.lan = lan
      self.stdout = Shell.OutputCapture()
      self.use_rawinput = False

    def log(self, s):
      self.stdout.write(s + "\n")

    def do_create(self, arg):
      'Create a job: hash manifest '
      self.lan.q.setJob(*(arg.split(maxsplit=1)))
      self.log(f"Added job")

    def do_acquire(self, arg):
      'Claim the job: hash'
      job = self.lan.q.acquireJob(arg)
      self.log(f"Acquired job '{arg}'")

    def do_release(self, arg):
      'Release the job: hash'
      job = self.lan.q.releaseJob(arg)
      self.log(f"Released job '{arg}'")
 
    def do_remove(self, arg):
      'Remove the job: hash'
      local_id = self.lan.q.removeJob(arg)
      self.log(f"Removed job '{arg}'")
 
    def do_status(self, arg):
      self.lan.q.printStatus()

    # ==== Metadata commands ====
    
    def do_syncpeer(self, arg):
      'Sync peer state: status run'
      status, runstr = arg.split(maxsplit=1)
      self.lan.q.syncPeer(status, json.loads(runstr))
      self.log(f"Synced peer state ({arg})")

    # ===== Database and inmemory (network queue) getters =====

    def do_peers(self, arg):
      'Print peers'
      for name, data in self.lan.q.getPeers().items():
          self.log(f"{name}: {data}")

    def do_jobs(self, arg):
      for job in self.lan.q.getJobs():
          self.log(str(job))

    def do_locks(self, arg):
      for (peer, locks) in self.lan.q.locks.getPeerLocks().items():
          self.log(f"{peer}: {locks}")


def main():
    logging.basicConfig(level=logging.DEBUG)
    import sys  
    import argparse

    from .lan_queue import LANPrintQueue
    parser = argparse.ArgumentParser(description='Start a network queue server')
    parser.add_argument('--ns', type=str)
    parser.add_argument('--addr', type=str)
    args = parser.parse_args()

    def on_update(q):
        print("<update>")

    lan = LANPrintQueue(args.ns, args.addr, None, on_update, logging.getLogger("queue"))
    sh = Shell()
    sh.attach(lan)
    while True:
      msg = input(">> ")
      sh.onecmd(msg)
      sys.stdout.write(sh.stdout.dump())

if __name__ == "__main__":
  main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/jobs.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10proto/jobs.proto\x12\x04jobs\"%\n\x04Lock\x12\x0c\n\x04peer\x18\x01 \x01(\t\x12\x0f\n\x07\x63reated\x18\x02 \x01(\x04\"Z\n\x03Job\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\t\x12\x18\n\x04lock\x18\x03 \x01(\x0b\x32\n.jobs.Lock\x12\x10\n\x08protocol\x18\x04 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\"\x10\n\x0eGetJobsRequest\"\'\n\rSetJobRequest\x12\x16\n\x03job\x18\x01 \x01(\x0b\x32\t.jobs.Job\"\x1e\n\x10\x44\x65leteJobRequest\x12\n\n\x02id\x18\x01 \x01(\t\"\x1f\n\x11\x41\x63quireJobRequest\x12\n\n\x02id\x18\x01 \x01(\t\"\x1f\n\x11ReleaseJobRequest\x12\n\n\x02id\x18\x01 \x01(\t\".\n\x13JobMutationResponse\x12\x17\n\x04jobs\x18\x01 \x03(\x0b\x32\t.jobs.JobB.Z,github.com/smartin015/peerprint/pubsub/protob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.jobs_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z,github.com/smartin015/peerprint/pubsub/proto'
  _LOCK._serialized_start=26
  _LOCK._serialized_end=63
  _JOB._serialized_start=65
  _JOB._serialized_end=155
  _GETJOBSREQUEST._serialized_start=157
  _GETJOBSREQUEST._serialized_end=173
  _SETJOBREQUEST._serialized_start=175
  _SETJOBREQUEST._serialized_end=214
  _DELETEJOBREQUEST._serialized_start=216
  _DELETEJOBREQUEST._serialized_end=246
  _ACQUIREJOBREQUEST._serialized_start=248
  _ACQUIREJOBREQUEST._serialized_end=279
  _RELEASEJOBREQUEST._serialized_start=281
  _RELEASEJOBREQUEST._serialized_end=312
  _JOBMUTATIONRESPONSE._serialized_start=314
  _JOBMUTATIONRESPONSE._serialized_end=360
# @@protoc_insertion_point(module_scope)



import unittest
import logging
from unittest.mock import MagicMock, call, patch, ANY
from .registry import IPFSRegistry
import peerprint.server.proto.peers_pb2 as ppb
import yaml
import json
from google.protobuf.json_format import MessageToJson


class TestRegistry(unittest.TestCase):
    def _writeReg(self, reg, dest):
        with open(dest, 'w') as f:
            yaml.dump(json.loads(MessageToJson(reg)), f)
            print("Dumped yaml to", dest)

    def testGetTrustedPeersAndRendezvous(self):
        reg = ppb.Registry(queues=[
            ppb.Queue(name="q1", trustedPeers=["p1_1", "p1_2"], rendezvous="r1"),
            ppb.Queue(name="q2", trustedPeers=["p2_1", "p2_2"], rendezvous="r2"),
            ])
        with patch('peerprint.wan.registry.IPFS') as fs:
            fs.fetch.side_effect = lambda cid, dest: self._writeReg(reg, dest)
            r = IPFSRegistry("testcid")
        
            self.assertEqual(r.get_rendezvous("q2"), reg.queues[1].rendezvous)
            self.assertEqual(r.get_trusted_peers("q1"), reg.queues[0].trustedPeers)




import unittest
from unittest.mock import patch
import logging
import tempfile
from pathlib import Path
from .filesharing import pack_job, unpack_job, packed_name, Fileshare, IPFSFileshare

class TestPackJob(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()
        p = Path(self.tempdir.name)
        self.pathnames = ['a.gcode', 'b.gcode', 'c.gcode']
        self.paths = dict([(n, p / n) for n in self.pathnames])
        for path in self.paths.values():
            path.touch()
        self.m = dict(sets=[dict(path=n) for n in self.pathnames])

    def tearDown(self):
        self.tempdir.cleanup()

    def test_pack_job_with_files(self):
        with tempfile.NamedTemporaryFile(suffix=".zip") as outpath:
            pack_job(self.m, self.paths, outpath.name)
            with tempfile.TemporaryDirectory() as td:
                result = unpack_job(outpath.name, td)
                result[0].pop("version")
                self.assertEqual(result[0], self.m)
                self.assertEqual([Path(p).name for p in result[1]], list(self.paths.keys()))

    def test_pack_job_hash_matching(self):
        with tempfile.NamedTemporaryFile(suffix=".zip") as tf1:
            with tempfile.NamedTemporaryFile(suffix=".zip") as tf2:
                h1 = pack_job(self.m, self.paths, tf1.name)
                h2 = pack_job(self.m, self.paths, tf2.name)
                self.assertEqual(h1, h2)

    def test_pack_job_hash_not_matching(self):
        with tempfile.NamedTemporaryFile(suffix=".zip") as tf1:
            with tempfile.NamedTemporaryFile(suffix=".zip") as tf2:
                h1 = pack_job(self.m, self.paths, tf1.name)
                self.m['sets'].pop(0)
                h2 = pack_job(self.m, dict(list(self.paths.items())[1:]), tf2.name)
                self.assertNotEqual(h1, h2)

    def test_throws_on_missing_file(self):
        self.m['sets'].append(dict(path="notexisting.gcode"))
        with tempfile.NamedTemporaryFile(suffix=".zip") as outpath:
            with self.assertRaises(ValueError):
                pack_job(self.m, self.paths, outpath.name)

class TestPackedName(unittest.TestCase):
    def testPacking(self):
        ts = 123
        for tc, want in [
            ("", "untitled.gjob"),
            ("hi", "hi.gjob"),
            ("!!**$$..", "xxxxxxxx.gjob"),
            ("space is cool", "space_is_cool.gjob"),
        ]:
            self.assertEqual(packed_name(tc, ''), want)

    def testPackingSuffix(self):
        class P:
            def __init__(self, path):
                self.base = path
                self.path = ''
                self.i = 0
            def __truediv__(self, other):
                self.path = self.base + other
                return self
            def __str__(self):
                return self.path
            def exists(self):
                self.i += 1
                if self.i < 5:
                    return True
        self.assertEqual(packed_name("test", P('/base/')), '/base/test (4).gjob') 

class TestFileshare(unittest.TestCase):
    def setUp(self):
        NUMFS = 2
        self.td = [tempfile.TemporaryDirectory() for i in range(NUMFS)]
        self.fs = [Fileshare("localhost:0", self.td[i].name, logging.getLogger(f"TestFileshare{i}")) for i in range(NUMFS)]
        for fs in self.fs:
            fs.connect()
        self.addr = [f"localhost:{fs.httpd.socket.getsockname()[1]}" for fs in self.fs]

    def tearDown(self):
        for td in self.td:
            td.cleanup()

    def testPostReceive(self):
        with tempfile.TemporaryDirectory() as td:
            p = Path(td) / 'a.gcode'
            DATA = "hello"
            with open(p, 'w') as f:
                f.write(DATA)
            m = dict(sets=[dict(path='a.gcode')])
            HASH = self.fs[0].post(m, {'a.gcode': p})
            dest = self.fs[1].fetch(self.addr[0], HASH, unpack=True)
            self.assertEqual(Path(dest).is_dir(), True)
            self.assertEqual((Path(dest) / 'a.gcode').exists(), True)
            with open(Path(dest) / 'a.gcode', 'r') as f:
                self.assertEqual(f.read(), DATA)


class TestIPFSFileshare(unittest.TestCase):
    @patch('peerprint.filesharing.IPFS')
    def testPost(self, tpfs):
        with tempfile.TemporaryDirectory() as td:
            tpfs.add.return_value = b"foo"
            fs = IPFSFileshare(td, logging.getLogger())
            ret = fs.post(dict(sets=[]), {})

            self.assertEqual(ret, "foo")
            self.assertTrue((Path(td) / "foo.gjob").exists())

    @patch("peerprint.filesharing.IPFS")
    def testFetch(self, tpfs):
        with tempfile.TemporaryDirectory() as td:
                tpfs.fetch.return_value = True
                fs = IPFSFileshare(td, logging.getLogger())
                ret = fs.fetch("foo", unpack=False, overwrite=False)
                self.assertEqual(ret, Path(td) / "foo.gjob")



import subprocess
import atexit
import signal
from dataclasses import dataclass, fields

@dataclass
class ServerProcessOpts():
    # See ../server/main.go for flag defs and defaults
    addr: str = None
    raftAddr: str = None
    raftPath: str = None
    rendezvous: str = None
    trustedPeers: str = None
    queue: str = None
    ipfs_server: str = None
    local: bool = None
    privkeyfile: str = None
    pubkeyfile: str = None
    connectTimeout: str = None
    zmq: str = None
    zmqpush: str = None
    zmqlog: str = None

    def render(self, binary_path):
        args = [binary_path]
        for field in fields(self):
            val = getattr(self, field.name)
            if val is not None:
                args.append(f"-{field.name}={val}")
        return args

class DependentProcess:
    def __init__(self):
        atexit.register(self.destroy)

    def _signal(self, sig, timeout=5):
        self._proc.send_signal(sig)
        try:
            self._proc.wait(timeout)
        except subprocess.TimeoutExpired:
            pass
        return self._proc.returncode is not None

    def destroy(self):
        atexit.unregister(self.destroy)

        if getattr(self, "_proc", None) is None or self._proc.returncode is not None:
            return
        self._logger.info(f"PID {self._proc.pid} (SIGINT)")
        if self._signal(signal.SIGINT):
            return

        self._logger.info(f"PID {self._proc.pid} (SIGKILL)")
        if self._signal(signal.SIGKILL):
            return

        self._logger.info(f"PID {self._proc.pid} (SIGTERM)")
        self._signal(signal.SIGTERM, timeout=None)

class IPFSDaemonProcess(DependentProcess):
    def __init__(self, logger):
        self._logger = logger
        if self.is_running():
            self._logger.warning("IPFS daemon already running; skipping daemon creation")
            self._proc = None
        else:
            self._proc = subprocess.Popen(["ipfs", "daemon", "--init", "--enable-pubsub-experiment"])
            self._logger.info("Launched IPFS daemon")
        super().__init__()

    def is_running(self):
        p = subprocess.Popen(["ipfs", "stats", "bw"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        p.communicate()
        return (p.returncode == 0)


class ServerProcess(DependentProcess):
    def __init__(self, opts, binary_path, logger):
        self._logger = logger
        args = opts.render(binary_path)
        self._proc = subprocess.Popen(args)
        self._logger.debug(f"Launch: {args}")

    def is_ready(self):
        return self._proc.returncode is None



import unittest
from unittest.mock import MagicMock, patch
from .ipfs import IPFS

class TestIPFS(unittest.TestCase):
    def setUp(self):
        with patch('peerprint.wan.ipfs.IPFSDaemonProcess') as dp:
            dp.is_running.return_value = True
            IPFS.start_daemon()

    def tearDown(self):
        IPFS.stop_daemon()

    def testAdd(self):
        with patch('peerprint.wan.ipfs.subprocess') as subp:
            subp.run.return_value = MagicMock(returncode=0, stdout="test_cid")
            self.assertEqual(IPFS.add("testpath"), "test_cid")
            subp.run.assert_called_once()
            self.assertEqual(subp.run.call_args[0][0][:2], ["ipfs", "add"])

    def testPin(self):
        with patch('peerprint.wan.ipfs.subprocess') as subp:
            subp.run.return_value = MagicMock(returncode=0)
            self.assertTrue(IPFS.pin("testpath"))
            subp.run.assert_called_once()
            self.assertEqual(subp.run.call_args[0][0][:3], ["ipfs", "pin", "add"])

    def testUnpin(self):
        with patch('peerprint.wan.ipfs.subprocess') as subp:
            subp.run.return_value = MagicMock(returncode=0)
            self.assertTrue(IPFS.unpin("testpath"))
            subp.run.assert_called_once()
            self.assertEqual(subp.run.call_args[0][0][:3], ["ipfs", "pin", "rm"])

    def testStat(self):
        with patch('peerprint.wan.ipfs.subprocess') as subp:
            subp.run.return_value = MagicMock(returncode=0, stdout="A:5\nB:10\nC:7".encode("utf8"))
            self.assertEqual(IPFS.stat("testpath"), dict(A=5, B=10, C=7))
            subp.run.assert_called_once()
            self.assertEqual(subp.run.call_args[0][0][:3], ["ipfs", "object", "stat"])

    def testFetch(self):
        with patch('peerprint.wan.ipfs.subprocess') as subp:
            subp.run.side_effect = [
                    MagicMock(returncode=0, stdout=f"CumulativeSize:{IPFS.MAX_CUMULATIVE_SIZE}".encode('utf8')),
                    MagicMock(returncode=0),
            ]
            self.assertTrue(IPFS.fetch("testcid", "testdest"))
            self.assertEqual(subp.run.call_args[0][0][:2], ["ipfs", "get"])

    def testFetchFailureTooLarge(self):
        with patch('peerprint.wan.ipfs.subprocess') as subp:
            subp.run.return_value = MagicMock(returncode=0, stdout=f"CumulativeSize:{IPFS.MAX_CUMULATIVE_SIZE+1}".encode('utf8'))
            with self.assertRaises(Exception):
                IPFS.fetch("testcid", "testdest")

    def testFetchStr(self):
        with patch('peerprint.wan.ipfs.subprocess') as subp:
            subp.run.return_value = MagicMock(returncode=0, stdout="test".encode('utf8'))
            self.assertEqual(IPFS.fetchStr("testcid"), "test")
            subp.run.assert_called_once()
            self.assertEqual(subp.run.call_args[0][0][:2], ["ipfs", "cat"])



import unittest
import queue
from unittest.mock import MagicMock, patch
from .comms import ZMQLogSink, ZMQCallError, ZMQClient, MessageUnpackError
import logging
import zmq
from google.protobuf.wrappers_pb2 import StringValue
from google.protobuf.message import Message
from google.protobuf.any_pb2 import Any
import peerprint.server.proto.state_pb2 as spb

logging.basicConfig(level=logging.DEBUG)


class TestZMQLogSink(unittest.TestCase):
    def testStreamLog(self):
        class DummySock:
            def __init__(self):
                self.closed = False

            def bind(self, addr):
                pass

            def recv(self):
                self.closed = True
                return "test\n".encode('utf8')

            def closed(self):
                return self.closed

        with patch('peerprint.wan.comms.zmq') as z:
            log = MagicMock()
            z.Context().socket.return_value = DummySock()
            l = ZMQLogSink("testaddr", log)
            log.info.assert_called_once_with("test")

    def testDestroy(self):
        with patch('peerprint.wan.comms.zmq') as z:
            self.l = ZMQLogSink("testaddr", MagicMock())
            self.l.destroy()
            z.Context().socket(zmq.PULL).destroy.assert_called_once()

class TestZMQClient(unittest.TestCase):
    def setUp(self):
        class DummySock:
            def __init__(self):
                self.closed = False
                self.q = queue.Queue()
                self.connect = MagicMock()
                self.bind = MagicMock()
                self.send = MagicMock()
                self.destroy = MagicMock()
            def recv(self):
                return self.q.get(block=True)

        with patch('peerprint.wan.comms.zmq') as z:
            self.notify = queue.Queue()
            self.cb = MagicMock(side_effect=lambda arg: self.notify.put(arg))
            self.pull = DummySock()
            self.req = DummySock()
            z.Context().socket.side_effect = [self.req, self.pull]
            self.c = ZMQClient("reqaddr", "pulladdr", self.cb, logging.getLogger())
            self.req.connect.assert_called_once()
            self.pull.bind.assert_called_once()

    def tearDown(self):
        if self.c is not None:
            self.pull.closed = True
            self.c.destroy()

    def testReceive(self):
        v = spb.Ok() # simple test message
        apb = Any()
        apb.Pack(v)
        self.pull.q.put(apb.SerializeToString())
        self.notify.get()
        self.cb.assert_called_once_with(v)

    def testReceiveUnknownMsg(self):
        v = StringValue(value="test")
        apb = Any()
        apb.Pack(v)
        self.pull.q.put(apb.SerializeToString())
        e = self.notify.get()
        self.cb.assert_called_once()
        self.assertTrue(str(e).startswith("Could not unpack"))
    
    def testCall(self):
        rep = spb.Ok() # simple test reply
        apb = Any()
        apb.Pack(rep)
        self.req.q.put(apb.SerializeToString())

        req = StringValue(value="test")
        apb2 = Any()
        apb2.Pack(req)

        got = self.c.call(req)
        self.req.send.assert_called_with(apb2.SerializeToString())
        self.assertEqual(got, rep)

    def testDestroy(self):
        self.c.destroy()
        self.c = None
        self.pull.destroy.assert_called_once()
        self.req.destroy.assert_called_once()



import socket
import random
import select
import time
import threading

from .version import __version__ as version

DISCOVERY_PORT = 37020

class P2PDiscovery:
  def __init__(self, namespace, advertise_addr, min_broadcast_pd=5, ttl=20):
    # Version is appended to the namespace to prevent different versions of peerprint from colliding
    self.namespace = f"{namespace}|{version}"
    self.ttl = ttl
    self.t = None
    self.min_broadcast_pd = min_broadcast_pd
    self.addr = advertise_addr
    self.sock = None
    self.host_timestamps = {}
  
  def _handlePing(self, host, ts):
    if host == self.addr:
      return # Don't count our own host in the list of peers
    # Filter all hosts that haven't contacted us in the TTL
    # print("ping", host, ts)
    if self.host_timestamps.get(host) is None:
      # print("_on_host_added", host)
      self._on_host_added(host)
      self.next_broadcast = 0 # Get the new host up to speed quickly
    self.host_timestamps[host] = ts
  
  def _prune(self, ts):
    removed = []
    for (k,v) in self.host_timestamps.items():
      if ts > v+self.ttl:
        removed.append(k)
    for k in removed:
      del self.host_timestamps[k]
      # print("_on_host_removed", k)
      self._on_host_removed(k)

  def _on_host_added(self, host):
    pass

  def _on_host_removed(self, host):
    pass

  def _on_startup_complete(self, results):
    pass

  def destroy(self):
    self.running = False

  def spin_async(self):
      self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) #, socket.IPPROTO_UDP)

      # Enable port reusage so we will be able to run multiple clients and servers on single (host, port). 
      # Do not use socket.SO_REUSEADDR except you using linux(kernel<3.9): goto https://stackoverflow.com/questions/14388706/how-do-so-reuseaddr-and-so-reuseport-differ for more information.
      # For linux hosts all sockets that want to share the same address and port combination must belong to processes that share the same effective user ID!
      # So, on linux(kernel>=3.9) you have to run multiple servers and clients under one user to share the same (host, port).
      # Thanks to @stevenreddie
      #
      # Note also that Windows does not have an SO_REUSEPORT variable - see https://github.com/smartin015/continuousprint/issues/154. Failover is required.
      self.sock.setsockopt(socket.SOL_SOCKET, getattr(socket, 'SO_REUSEPORT', socket.SO_REUSEADDR), 1)
      self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
      self.sock.bind(("", DISCOVERY_PORT))
      self.sock.settimeout(0.2)
      self.t = threading.Thread(target=self.spin, daemon=True)
      self.t.start()

  def spin(self):
    # print("spin start")
    startup = time.time()
    self.running = True
    while self.running:
      ts = time.time()

      # We should have received all hosts actively broadcasting on the network basically 
      # instantaneously with rebroadcasts, but wait a bit just in case
      if startup is not None and ts > startup+self.min_broadcast_pd/2:
        self._on_startup_complete(self.host_timestamps)
        startup = None

      self.sock.sendto(f"{self.namespace}|{self.addr}".encode(), ('<broadcast>', DISCOVERY_PORT))
      self.next_broadcast = ts + self.min_broadcast_pd * (1 + random.random())
      self._prune(ts)
      while ts < self.next_broadcast:
        ts = time.time()
        read_sockets, write_sockets, error_sockets = select.select([self.sock], [], [], 0.2)
        if len(read_sockets) > 0:
          #incoming message from remote server
          data = read_sockets[0].recv(4096)
          if data.startswith(self.namespace.encode()):
            data = data.decode("utf8").split("|")[-1]
            self._handlePing(data, ts)
    # print("spin end")

if __name__ == "__main__":
  import sys
  
  class BasicDiscovery(P2PDiscovery):
    def _on_host_added(self, host):
      print("ADD:", host)

    def _on_host_removed(self, host):
      print("RM:", host)

    def _on_startup_complete(self, results):
      print("START:", results)

  if len(sys.argv) != 3:
    sys.stderr.write("Usage: python3 discover.py [namespace] [address:port]\n")
    sys.exit(1)

  b = BasicDiscovery(sys.argv[1], sys.argv[2])
  print("Starting discovery")
  b.spin_async()
  input()
  




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/state.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from peerprint.server.proto import jobs_pb2 as proto_dot_jobs__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11proto/state.proto\x12\x05state\x1a\x10proto/jobs.proto\"e\n\x05State\x12$\n\x04jobs\x18\x01 \x03(\x0b\x32\x16.state.State.JobsEntry\x1a\x36\n\tJobsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.jobs.Job:\x02\x38\x01\"\x17\n\x05\x45rror\x12\x0e\n\x06status\x18\x02 \x01(\t\"\x04\n\x02OkB.Z,github.com/smartin015/peerprint/pubsub/protob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.state_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z,github.com/smartin015/peerprint/pubsub/proto'
  _STATE_JOBSENTRY._options = None
  _STATE_JOBSENTRY._serialized_options = b'8\001'
  _STATE._serialized_start=46
  _STATE._serialized_end=147
  _STATE_JOBSENTRY._serialized_start=93
  _STATE_JOBSENTRY._serialized_end=147
  _ERROR._serialized_start=149
  _ERROR._serialized_end=172
  _OK._serialized_start=174
  _OK._serialized_end=178
# @@protoc_insertion_point(module_scope)



from pysyncobj import SyncObjConsumer, replicated
from collections import defaultdict
import threading
import weakref
import time
import socket
import os

# See https://github.com/bakwc/PySyncObj/blob/master/pysyncobj/batteries.py for original
# ReplDict implementation.
# This dict also maintains an ordered index of objects
class CPOrderedReplDict(SyncObjConsumer):
    def __init__(self, cb):
        self.cb = cb

        # All non-synced attributes must occur BEFORE call to super()
        super().__init__()
        self.__data = {}

        # Indices for ordering queue items
        self.__before = {}
        self.__after = {}
        self.__first = None
        self.__last = None

    def _setitem_impl(self, key, value):
        prev = self.__data.get(key, None)
        self.__data[key] = value
        if prev is None: # Insert into order index
            self._link(key, self.__last)
        self.cb(prev, value)

    @replicated
    def __setitem__(self, key, value):
        return self._setitem_impl(key, value)

    def _pop_impl(self, key, default=None):
        if key not in self.__data:
            return None
        val = self.__data.pop(key, default)
        self._unlink(key)
        self.cb(val, None)
        return val

    @replicated
    def pop(self, key, default=None):
        return self._pop_impl(key, default)

    @replicated
    def mv(self, key, after):
        return self._mv_impl(key, after)

    def _unlink(self, key):
        # Unlink from current position
        prev = self.__before[key]
        nxt = self.__after[key]

        # Point prev to nxt
        if nxt is not None:
            self.__before[nxt] = prev
        if prev is not None:
            self.__after[prev] = nxt

        # Update first & last 
        if self.__before[key] is None:
            self.__first = self.__after[key]
        if self.__after[key] is None:
            self.__last = self.__before[key]

        # Remove unused keys
        del self.__before[key]
        del self.__after[key]

    def _link(self, key, after):
        # Invariant: key does not exist in the linked list
        assert key not in self.__before
        if after is None: # Send to front
            self.__after[key] = self.__first
            self.__before[key] = None
            self.__first = key
            if self.__last is None: # We are front and end of queue
                self.__last = key
        else: # Add to middle/end
            anxt = self.__after[after]

            # Make links pre-key
            self.__before[key] = after
            self.__after[after] = key

            # Make links post-key
            self.__before[anxt] = key
            self.__after[key] = anxt

            # Update end if needed
            if anxt is None:
                self.__last = key
    
    def _mv_impl(self, key, after):
        if after is not None and after not in self.__before:
            raise KeyError(after)
        self._unlink(key)
        self._link(key, after)

    def __getitem__(self, key):
        return self.__data[key]

    def get(self, key, default=None):
        return self.__data.get(key, default)

    def set(self, key, value, **kwargs):
        # Duplicated from __setitem__ to allow for passing decorator args (sync, timeout, callback)
        self.__setitem__(key, value, **kwargs)

    def __len__(self):
        return len(self.__data)

    def __contains__(self, key):
        return key in self.__data

    def keys(self):
        return self.__data.keys()

    def values(self):
        return self.__data.values()

    def items(self):
        return self.__data.items()

    def ordered_items(self):
        nxt = self.__first
        while nxt is not None:
            yield (nxt, self.__data.get(nxt))
            nxt = self.__after.get(nxt)

class _ReplLockManagerImpl(SyncObjConsumer):
    def __init__(self, autoUnlockTime, cb):
        self.cb = cb
        # All non-synced attributes must occur BEFORE call to super()
        super(_ReplLockManagerImpl, self).__init__()
        self.__locks = {}
        self.__autoUnlockTime = autoUnlockTime

    @replicated
    def acquire(self, lockID, clientID, currentTime):
        existingLock = self.__locks.get(lockID, None)
        # Auto-unlock old lock
        if existingLock is not None:
            if currentTime - existingLock[1] > self.__autoUnlockTime:
                existingLock = None
            else:
                existingLock = existingLock[0] # Unwrap to get client ID
        # Acquire lock if possible
        if existingLock is None or existingLock == clientID:
            self.__locks[lockID] = (clientID, currentTime)
            self.cb((lockID, existingLock), (lockID, clientID))
            return True
        # Lock already acquired by someone else
        return False

    @replicated
    def prolongate(self, clientID, currentTime):
        for lockID in list(self.__locks):
            lockClientID, lockTime = self.__locks[lockID]

            if currentTime - lockTime > self.__autoUnlockTime:
                del self.__locks[lockID]
                self.cb((lockID, lockClientID), None)
                continue

            if lockClientID == clientID:
                self.__locks[lockID] = (clientID, currentTime)

    @replicated
    def release(self, lockID, clientID):
        existingLock = self.__locks.get(lockID, None)
        if existingLock is not None and existingLock[0] == clientID:
            del self.__locks[lockID]
            self.cb((lockID, clientID), None)

    def isAcquired(self, lockID, clientID, currentTime):
        existingLock = self.__locks.get(lockID, None)
        if existingLock is not None:
            if existingLock[0] == clientID:
                if currentTime - existingLock[1] < self.__autoUnlockTime:
                    return True
        return False


class CPReplLockManager(object):

    def __init__(self, autoUnlockTime, selfID = None, cb = None):
        self.__lockImpl = _ReplLockManagerImpl(autoUnlockTime, cb)
        if selfID is None:
            selfID = '%s:%d:%d' % (socket.gethostname(), os.getpid(), id(self))
        self.__selfID = selfID
        self.__autoUnlockTime = autoUnlockTime
        self.__mainThread = threading.current_thread()
        self.__initialised = threading.Event()
        self.__destroying = False
        self.__lastProlongateTime = 0
        self.__thread = threading.Thread(target=CPReplLockManager._autoAcquireThread, args=(weakref.proxy(self),))
        self.__thread.start()
        while not self.__initialised.is_set():
            pass

    def _consumer(self):
        return self.__lockImpl

    def destroy(self):
        self.__destroying = True

    def _autoAcquireThread(self):
        self.__initialised.set()
        try:
            while True:
                if not self.__mainThread.is_alive():
                    break
                if self.__destroying:
                    break
                time.sleep(0.1)
                if time.time() - self.__lastProlongateTime < float(self.__autoUnlockTime) / 4.0:
                    continue
                syncObj = self.__lockImpl._syncObj
                if syncObj is None:
                    continue
                if syncObj._getLeader() is not None:
                    self.__lastProlongateTime = time.time()
                    self.__lockImpl.prolongate(self.__selfID, time.time())
        except ReferenceError:
            pass

    def tryAcquire(self, lockID, callback=None, sync=False, timeout=None):
        attemptTime = time.time()
        if sync:
            acquireRes = self.__lockImpl.acquire(lockID, self.__selfID, attemptTime, callback=callback, sync=sync, timeout=timeout)
            acquireTime = time.time()
            if acquireRes:
                if acquireTime - attemptTime > self.__autoUnlockTime / 2.0:
                    acquireRes = False
                    self.__lockImpl.release(lockID, self.__selfID, sync=sync)
            return acquireRes

        def asyncCallback(acquireRes, errCode):
            if acquireRes:
                acquireTime = time.time()
                if acquireTime - attemptTime > self.__autoUnlockTime / 2.0:
                    acquireRes = False
                    self.__lockImpl.release(lockID, self.__selfID, sync=False)
            callback(acquireRes, errCode)

        self.__lockImpl.acquire(lockID, self.__selfID, attemptTime, callback=asyncCallback, sync=sync, timeout=timeout)

    def isAcquired(self, lockID):
        return self.__lockImpl.isAcquired(lockID, self.__selfID, time.time())

    def release(self, lockID, callback=None, sync=False, timeout=None):
        self.__lockImpl.release(lockID, self.__selfID, callback=callback, sync=sync, timeout=timeout)

    def notAcquired(self, lockID: str) -> bool:
        """Check if lock is open / not acquired by anyone"""
        existingLock = self._consumer()._ReplLockManagerImpl__locks.get(lockID)
        if existingLock is not None:
            return time.time() - existingLock[1] > self._CPReplLockManager__autoUnlockTime
        return True

    def getPeerLocks(self) -> dict:
        result = defaultdict(list)
        now = time.time()
        for (lid, val) in self._consumer()._ReplLockManagerImpl__locks.items():
            (peer, acquired_at) = val
            if time.time() - acquired_at < self._CPReplLockManager__autoUnlockTime:
                result[peer].append(lid)
        return result




import subprocess
import logging
from .proc import IPFSDaemonProcess

proc = None

class IPFS:
    # Current python libraries designed for IPFS access are 5+ versions behind in compatibility
    # so we simply subprocess the ipfs-go CLI here.

    MAX_CUMULATIVE_SIZE = 10*1024*1024 # 10 MiB

    @classmethod
    def start_daemon(self):
        global proc
        if proc != None:
            return proc
        proc = IPFSDaemonProcess(logging.getLogger())
        return proc

    @classmethod 
    def stop_daemon(self):
        global proc
        if proc != None:
            proc.destroy()
            proc = None

    @classmethod
    def check(self):
        global proc
        if proc is None:
            raise Exception("IPFS daemon not started")
        if not proc.is_running():
            raise Exception("IPFS daemon not running")
        return

    @classmethod
    def add(self, path):
        IPFS.check()
        # Invoke ipfs CLI, add file, and return its hash
        # TODO check file size before adding so it doesn't get rejected out of hand
        result = subprocess.run(["ipfs", "add", "--cid-version=1", "-q", path], capture_output=True)
        if result.returncode != 0:
            raise Exception(f"add_file result {result.returncode}: {result.stderr}")
        return result.stdout.strip()

    @classmethod
    def pin(self, hash_):
        IPFS.check()
        result = subprocess.run(["ipfs", "pin", "add", hash_])
        return result.returncode == 0

    @classmethod
    def unpin(self, hash_):
        IPFS.check()
        result = subprocess.run(["ipfs", "pin", "rm", hash_])
        return result.returncode == 0

    @classmethod 
    def stat(self, hash_):
        IPFS.check()
        result = subprocess.run(["ipfs", "object", "stat", hash_], capture_output=True)
        if result.returncode != 0:
            raise Exception(f"stat result {result.returncode}: {result.stderr}")
        out = result.stdout.decode('utf8')
        result = dict()
        for line in out.split('\n'):
            if line.strip() == "":
                continue
            k, v = line.split(':', 1)
            result[k.strip()] = int(v)
        return result

    @classmethod
    def fetch(self, hash_, dest):
        IPFS.check()
        csz = self.stat(hash_)['CumulativeSize'] 
        if csz > self.MAX_CUMULATIVE_SIZE:
            raise Exception(f"IPFS: {hash_} cumulative size {csz} exceeds limit ({self.MAX_CUMULATIVE_SIZE})")

        result = subprocess.run(["ipfs", "get", f"--output={dest}", hash_])
        return result.returncode == 0

    @classmethod
    def fetchStr(self, hash_) -> str:
        IPFS.check()
        result = subprocess.run(["ipfs", "cat", hash_], capture_output=True)
        if result.returncode != 0:
            raise Exception(f"fetchStr result {result.returncode}: {result.stderr}")
        return result.stdout.decode('utf8').strip()


if __name__ == '__main__':
    result = IPFS.add("test.txt")
    print(result)
    print(IPFS.pin(result))
    print(IPFS.unpin(result))
    print(IPFS.fetch(result, "downloaded.txt"))



import unittest
from unittest.mock import MagicMock
from .sync_objects import CPOrderedReplDict

# Implement CPOrderedReplDict's abstract _item_changed function
# and also override replication on mutation methods
class TestReplDict(CPOrderedReplDict):
    def __init__(self, cb):
        super().__init__(cb)

    def _item_changed(self, prev, nxt):
        return prev != nxt

    def __setitem__(self, k, v, **kwargs):
        return self._setitem_impl(k,v)

    def pop(self, k,d=None):
        return self._pop_impl(k,d)

    def mv(self, key, after):
        return self._mv_impl(key, after)

class TestCPOrderedReplDictEmpty(unittest.TestCase):
    def setUp(self):
        self.m = MagicMock()
        self.d = TestReplDict(self.m)

    def testAddItemsInOrder(self):
        for a in ['d','c']:
            self.d[a] = a

        self.assertEqual(list(self.d.ordered_items()), [
            ('d', 'd'),
            ('c', 'c'),
        ])

    def testPopEmpty(self):
        self.assertEqual(self.d.pop('a'), None)

    def testMoveSingleItemInQueue(self):
        self.d['1'] = '1'
        self.d.mv('1', None)
        self.assertEqual(list(self.d.ordered_items()), [('1', '1')])

class TestCPOrderedReplDictWithItems(unittest.TestCase):
    def setUp(self):
        self.m = MagicMock()
        self.d = TestReplDict(self.m)
        for a in ['1','2','3','4']:
            self.d[a] = a

    def _orderedEquals(self, result):
        return self.assertEqual(
            [i[0] for i in list(self.d.ordered_items())],
            result)

    def testPopMiddle(self):
        self.d.pop('2')
        self._orderedEquals(['1','3','4'])

    def testPopLast(self):
        self.d.pop('4')
        self._orderedEquals(['1','2','3'])

    def testPopFirst(self):
        self.d.pop('1')
        self._orderedEquals(['2','3','4'])

    def testMoveExchangeMiddle(self):
        self.d.mv('2', '3')
        self._orderedEquals(['1','3','2','4'])

        # Reverse direction
        self.d.mv('2', '1')
        self._orderedEquals(['1','2','3','4'])

    def testMoveLastToFirst(self):
        self.d.mv('4', None)
        self._orderedEquals(['4','1','2','3'])

    def testMoveFirstToLast(self):
        self.d.mv('1', '4')
        self._orderedEquals(['2','3','4','1'])

    def testMoveLastToMiddle(self):
        self.d.mv('4', '2')
        self._orderedEquals(['1','2','4','3'])

    def testMoveBadAfterCausesNoChange(self):
        with self.assertRaises(KeyError):
            self.d.mv('4', '80')
        self._orderedEquals(['1','2','3','4'])

    def testMoveBadKeyCausesNoChange(self):
        with self.assertRaises(KeyError):
            self.d.mv('80', '2')
        self._orderedEquals(['1','2','3','4'])



import hashlib
import requests
import zipfile
import shutil
import os
import json
import time
import re
import tempfile
from .wan.ipfs import IPFS
from abc import ABC, abstractmethod
from pathlib import Path
from .version import __version__ as version
import http.server
import socketserver
import threading

try:
    import zlib
    compression = zipfile.ZIP_DEFLATED
except:
    compression = zipfile.ZIP_STORED

# https://stackoverflow.com/a/3431838
def _content_hash(path) -> str:
    h = hashlib.sha256()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            h.update(chunk)
    return h.hexdigest()


# Inspired by https://stackoverflow.com/a/1007615
def packed_name(s, basedir):
    if isinstance(basedir, str):
        basedir = Path(basedir)
    if s.strip() == "":
        s = "untitled"

    # Replace all non-word characters (everything except numbers and letters)
    s = re.sub(r"[^\w\s]", 'x', s)
    # Replace all runs of whitespace with underscore
    s = re.sub(r"\s+", '_', s)
    
    # Name suffix inspired by https://stackoverflow.com/a/57896232
    path = basedir / f"{s}.gjob" 
    counter = 1
    while path.exists():
        path = basedir / f"{s} ({counter}).gjob"
        counter += 1
    return str(path)


def pack_job(manifest: dict, filepaths: dict, dest: str):
    # TODO validation - ensure the correct files are available given the manifest object)
    zf = zipfile.ZipFile(dest, mode='w')

    # Sanitize short paths
    filepaths = dict([(short.split("/")[-1], full) for (short, full) in filepaths.items()])

    # Strip off paths in manifest (paths are stripped in zip file as well)
    # and remove unnecessary state/identity fields.
    for s in manifest["sets"]:
        s["path"] = s["path"].split("/")[-1]
        if filepaths.get(s["path"]) is None:
            raise ValueError(f"Job contains set with path={s['path']}, but filepaths has no matching short name")
        for k in ("sd"):
            # Note: we leave ID and rank around here as it's useful for ordering/referral to set items
            # Also leave "remaining" so that dragging between local and LAN queues is consistent.
            s.pop(k, None)
    for k in ("acquired", "id", "queue"):
        manifest.pop(k, None)


    try:
        for (shortpath, fullpath) in filepaths.items():
            zf.write(fullpath, arcname=shortpath, compress_type=compression)
        zf.writestr("manifest.json", json.dumps(dict(**manifest, version=version)))
    finally:
        zf.close()
    
    return _content_hash(dest)


def unpack_job(path, outdir):
    with zipfile.ZipFile(path, mode='r') as zf:
        zf.extractall(path=outdir)
        with open(Path(outdir) / 'manifest.json', 'r') as f:
            manifest = json.loads(f.read())
        return manifest, [n for n in zf.namelist() if Path(n).name != 'manifest.json']

class IPFSFileshare():
    def __init__(self, basedir, logger):
        self.basedir = basedir
        self._logger = logger
        os.makedirs(basedir, exist_ok=True)
        self.proc = IPFS.start_daemon()

    def post(self, manifest: dict, filepaths: dict) -> str:
        with tempfile.NamedTemporaryFile(suffix='.gjob', dir=self.basedir, delete=False) as tf:
            hash_ = pack_job(manifest, filepaths, tf.name)
            ipfs_cid = IPFS.add(tf.name).decode('utf8')
            dest = Path(self.basedir) / f"{ipfs_cid}.gjob"
            os.rename(tf.name, dest)
            self._logger.info(f"Packed and posted job to {dest} - IPFS id {ipfs_cid}")
            return ipfs_cid

    def fetch(self, hash_:str, unpack=False, overwrite=False) -> str:
        # Get the equivalent path on disk
        name = f"{hash_}.gjob"
        dest = Path(self.basedir) / name
        if dest.exists() and not overwrite:
            self._logger.debug("File already exists - using that one")
        else:
            if not IPFS.fetch(hash_, dest):
                raise Exception("Failed to fetch file {hash_}")

        if unpack:
            dest_dir = Path(self.basedir) / hash_
            if not dest_dir.exists() or overwrite:
                unpack_job(dest, dest_dir)
            return dest_dir
        else:
            return dest


class FileshareServer(socketserver.TCPServer):
    allow_reuse_address = True

class Fileshare:
    def __init__(self, addr, basedir, logger):
        (host, port) = addr.split(":")
        self.host = host
        self.port = int(port)
        self.basedir = basedir
        self.t = None
        self._logger = logger
        os.makedirs(basedir, exist_ok=True)

    def connect(self, testing=False):
        if testing:
            return
        basedir = self.basedir
        class FileshareRequestHandler(http.server.SimpleHTTPRequestHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, directory=basedir, **kwargs)

        self.httpd = FileshareServer((self.host, self.port), FileshareRequestHandler)
        self.httpd.allow_reuse_address = True
        self.t = threading.Thread(target=self.httpd.serve_forever, daemon=True)
        self.t.start()

        # Re-assign host & port since some formats (e.g. *:0) auto-assign port
        (self.host, self.port) = self.httpd.socket.getsockname()

    def destroy(self):
        if self.httpd is not None:
            self._logger.info("Fileshare shutting down")
            self.httpd.shutdown()
            self.httpd.server_close()
            self._logger.info("Server shut down")

    def post(self, manifest: dict, filepaths: dict) -> str:
        # We must first write to a temp file so we can calculate the hash and use it
        # as the destination file name
        # Note that basedir must be used when creating the file, as it otherwise defaults to /tmp
        # which is in-memory and causes os.rename to fail due to cross-device linking.
        with tempfile.NamedTemporaryFile(suffix='.gjob', dir=self.basedir, delete=False) as tf:
            hash_ = pack_job(manifest, filepaths, tf.name)
            dest = Path(self.basedir) / f"{hash_}.gjob"
            os.rename(tf.name, dest)
            self._logger.info(f"Packed and posted job to {dest}")
            return hash_

    def fetch(self, peer:str, hash_:str, unpack=False, overwrite=False) -> str:
        # Get the equivalent path on disk
        written = 0
        name = f"{hash_}.gjob"
        url = f"http://{peer}/{name}"
        dest = Path(self.basedir) / name
        self._logger.debug(f"HTTP GET {url} -> {dest}")

        if dest.exists() and not overwrite:
            self._logger.debug("File already exists - using that one")
        else:
            with requests.get(url, stream=True) as r:
              r.raise_for_status()
              with open(dest, 'wb') as f:
                  for chunk in r.iter_content(chunk_size=8192):
                      f.write(chunk)
                      written += len(chunk)
            self._logger.debug(f"Wrote {written}B to {dest}")

        if unpack:
            dest_dir = Path(self.basedir) / hash_
            if not dest_dir.exists() or overwrite:
                unpack_job(dest, dest_dir)
            return dest_dir
        else:
            return dest



from .wan_queue import PeerPrintQueue
from .proc import ServerProcessOpts
from .registry import FileRegistry
import subprocess
from threading import Lock
import tempfile
import logging
import time
import sys
import os
import json

print(__file__)

logging.basicConfig(level=logging.DEBUG)
TEST_JOB = "asdfghjk"
OP_TIMEOUT = 3

if len(sys.argv) != 2:
    raise Exception(f"Usage: {sys.argv[0]} <path_to_peerprint_server>")
server_path = sys.argv[1]

class JSONCodec():
    @classmethod
    def encode(self, manifest):
        return (json.dumps(manifest).encode("utf8"), "json")

    @classmethod
    def decode(self, data, protocol):
        assert protocol=="json"
        return json.loads(data.decode("utf8"))

lock = Lock()
lock.acquire()

def on_update(changetype, prev, nxt):
    print("on_update", prev, "->", nxt)
    try:
        lock.release()
    except RuntimeError:
        pass

def do_get(fn):
    fn()
    if not lock.acquire(timeout=OP_TIMEOUT):
        raise Exception(f"Failed to provoke a state change after {OP_TIMEOUT}s")
    return q1.getJobs()

def make_queue(tdir, idx):
    reg = FileRegistry(os.path.join(tdir, "registry.yaml"))
    qname = reg.get_queue_names()[0]
    tpeers = reg.get_trusted_peers(qname)
    ppq = PeerPrintQueue(ServerProcessOpts(
            rendezvous=reg.get_rendezvous(qname),
            trustedPeers=",".join(tpeers),
            local=True,
            raftPath=f"./peer{idx}.raft",
            privkeyfile=os.path.join(tdir, f"trusted_peer_{idx+1}.priv"),
            pubkeyfile=os.path.join(tdir, f"trusted_peer_{idx+1}.pub"),
    ), tpeers[idx], JSONCodec, server_path, on_update if idx == 1 else None, logging.getLogger(f"q{idx}"))
    logging.info(f"Starting connection ({idx})")
    ppq.connect()
    return ppq

def make_config(tdir, npeers):
    assert subprocess.run([server_path, "generate_registry", str(npeers), tdir]).returncode == 0

def main(q0, q1):
    print("Waiting for queues to be ready...")
    while not q0.is_ready() or not q1.is_ready():
        time.sleep(1)

    print("Waiting for state update")
    lock.acquire()

    print("Querying peers, jobs")
    rep = q1.getPeers()
    print(f"q1 {rep.peer_estimate} peers (variance {rep.variance} from a sample of {len(rep.sample)})")
    rep = q1.getJobs()
    print(f"q1: {len(rep)} jobs: {rep}")

    print("q0: uploading a dummy job")
    rep = do_get(lambda: q0.setJob(TEST_JOB, dict(man="ifest")))
    print(f"q1: {len(rep)} jobs: {rep}")
    if len(rep) != 1:
        raise Exception(f"Expected 1 job in queue, got {len(rep)}")

    print("q0: acquiring the job")
    rep = do_get(lambda: q0.acquireJob(TEST_JOB))
    print("q0: job is now", rep[TEST_JOB])
    if not rep[TEST_JOB]['acquired'] or rep[TEST_JOB]['acquired_by_'] != q0.get_id():
        raise Exception(f"Expected job {TEST_JOB} to be acquired by q0 (id {q0.get_id()}, got: {rep[TEST_JOB]}")

    print("q0: releasing the job")
    rep = do_get(lambda: q0.releaseJob(TEST_JOB))
    if rep[TEST_JOB].get('acquired', None) is not None:
        raise Exception(f"Expected job {TEST_JOB} to be not acquired, but it is")

    print("q0: removing the job")
    rep = do_get(lambda: q0.removeJob(TEST_JOB))
    if rep.get(TEST_JOB) != None:
        raise Exception(f"Expected job ID {TEST_JOB} to be removed, but it still exists")

    print("SUCCESS - all tasks achieved")


if __name__ == "__main__":
    with tempfile.TemporaryDirectory() as tdir: 
        print("Creating config files")
        make_config(tdir, 2)
    
        print("Constructing queues")
        q0 = make_queue(tdir, 0)
        q1 = make_queue(tdir, 1)

        print("Running demo")
        main(q0, q1)




from .ipfs import IPFS
from pathlib import Path
import yaml
import json
from google.protobuf import json_format
import peerprint.server.proto.peers_pb2 as ppb
import os
import tempfile
import time

class Registry:
    def __init__(self):
        self.registry = None
        self.qmap = dict()

    def ready(self):
        return self.proc.is_running()
 
    def _fetch(self):
        raise NotImplemented

    def _load_registry(self):
        data = self._fetch()

        # Protobufs can serialize to binary or json, but registries
        # are written in a YAML format. We serialize to JSON before
        # deserializing to proto - inefficient, but convenient.
        ydata = yaml.safe_load(data)
        reg = ppb.Registry()
        json_format.Parse(json.dumps(ydata), reg)
        self.registry = reg
        self.qmap = dict([(q.name, q) for q in self.registry.queues])
        assert self.registry is not None

    def _get_queue(self, queue: str):
        if self.registry is None:
            self._load_registry()
        q = self.qmap.get(queue)
        if q is None:
            raise Exception(f"Queue '{queue}' not found in registry - candidates: {self.qmap.keys()}")
        return q

    def get_queue_names(self):
        if self.registry is None:
            self._load_registry()
        return list(self.qmap.keys())

    def get_trusted_peers(self, queue: str):
        q = self._get_queue(queue)
        return q.trustedPeers

    def get_rendezvous(self, queue: str):
        q = self._get_queue(queue)
        return q.rendezvous


class IPFSRegistry(Registry):
    def __init__(self, cid: str):
        super().__init__()
        self.cid = cid
        self.proc = IPFS.start_daemon()
        time.sleep(0.5)

    def _fetch(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            dest = Path(temp_dir) / "registry.yaml"
            IPFS.fetch(self.cid, dest)
            try:
                with open(dest, 'r') as f:
                    data = f.read()
            finally:
                os.remove(dest)
        return data

class FileRegistry(Registry):
    def __init__(self, path: str):
        super().__init__()
        self.path = path

    def _fetch(self):
        with open(self.path, 'r') as f:
            return f.read()

if __name__ == "__main__":
    import time
    import sys
    if len(sys.argv) != 2:
        # CID e.g. QmZQ4bLHRCrcmJUnTbY7updR6chfvumbaEx6cCya3chz9n
        raise Exception(f"Usage: {sys.argv[0]} <IPFS CID>")
    r = Registry(sys.argv[1])

    print("connecting")
    while not r.ready():
        print(".")
        time.sleep(1)

    tp = r.get_trusted_peers("Test queue")
    rend = r.get_rendezvous("Test queue")
    print("\n\n\n=========================== Queue data: ============================\n")
    print("\t- Trusted peers:", tp)
    print("\t- Rendezvous:", rend)
    print("\n======================================================================")




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/peers.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11proto/peers.proto\x12\x03rpc\"Z\n\x05Queue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x65sc\x18\x02 \x01(\t\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x12\n\nrendezvous\x18\x04 \x01(\t\x12\x14\n\x0ctrustedPeers\x18\x05 \x03(\t\"D\n\x08Registry\x12\x0f\n\x07\x63reated\x18\x01 \x01(\x04\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\x1a\n\x06queues\x18\x03 \x03(\x0b\x32\n.rpc.Queue\"s\n\nPeerStatus\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05topic\x18\x02 \x01(\t\x12\x0e\n\x06leader\x18\x03 \x01(\t\x12\x1b\n\x04type\x18\x04 \x01(\x0e\x32\r.rpc.PeerType\x12\x1d\n\x05state\x18\x05 \x01(\x0e\x32\x0e.rpc.PeerState\"\x14\n\x06Leader\x12\n\n\x02id\x18\x01 \x01(\t\"X\n\x0cPeersSummary\x12\x15\n\rpeer_estimate\x18\x01 \x01(\x03\x12\x10\n\x08variance\x18\x02 \x01(\x01\x12\x1f\n\x06sample\x18\x03 \x03(\x0b\x32\x0f.rpc.PeerStatus\"6\n\x10PollPeersRequest\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x13\n\x0bprobability\x18\x02 \x01(\x01\"4\n\x11PollPeersResponse\x12\x1f\n\x06status\x18\x01 \x01(\x0b\x32\x0f.rpc.PeerStatus\"\x13\n\x11\x41ssignmentRequest\"%\n\x08\x41\x64\x64rInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05\x61\x64\x64rs\x18\x02 \x03(\t\"4\n\x10RaftAddrsRequest\x12 \n\taddr_info\x18\x01 \x01(\x0b\x32\r.rpc.AddrInfo\"5\n\x11RaftAddrsResponse\x12 \n\taddr_info\x18\x01 \x01(\x0b\x32\r.rpc.AddrInfo\"_\n\x12\x41ssignmentResponse\x12\r\n\x05topic\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12\x11\n\tleader_id\x18\x03 \x01(\t\x12\x1b\n\x04type\x18\x04 \x01(\x0e\x32\r.rpc.PeerType\"\x1f\n\x11NewLeaderResponse\x12\n\n\x02id\x18\x01 \x01(\t*>\n\x08PeerType\x12\x15\n\x11UNKNOWN_PEER_TYPE\x10\x00\x12\r\n\tELECTABLE\x10\x02\x12\x0c\n\x08LISTENER\x10\x03*C\n\tPeerState\x12\x16\n\x12UNKNOWN_PEER_STATE\x10\x00\x12\x08\n\x04\x42USY\x10\x01\x12\x08\n\x04IDLE\x10\x02\x12\n\n\x06PAUSED\x10\x03\x42.Z,github.com/smartin015/peerprint/pubsub/protob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.peers_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z,github.com/smartin015/peerprint/pubsub/proto'
  _PEERTYPE._serialized_start=826
  _PEERTYPE._serialized_end=888
  _PEERSTATE._serialized_start=890
  _PEERSTATE._serialized_end=957
  _QUEUE._serialized_start=26
  _QUEUE._serialized_end=116
  _REGISTRY._serialized_start=118
  _REGISTRY._serialized_end=186
  _PEERSTATUS._serialized_start=188
  _PEERSTATUS._serialized_end=303
  _LEADER._serialized_start=305
  _LEADER._serialized_end=325
  _PEERSSUMMARY._serialized_start=327
  _PEERSSUMMARY._serialized_end=415
  _POLLPEERSREQUEST._serialized_start=417
  _POLLPEERSREQUEST._serialized_end=471
  _POLLPEERSRESPONSE._serialized_start=473
  _POLLPEERSRESPONSE._serialized_end=525
  _ASSIGNMENTREQUEST._serialized_start=527
  _ASSIGNMENTREQUEST._serialized_end=546
  _ADDRINFO._serialized_start=548
  _ADDRINFO._serialized_end=585
  _RAFTADDRSREQUEST._serialized_start=587
  _RAFTADDRSREQUEST._serialized_end=639
  _RAFTADDRSRESPONSE._serialized_start=641
  _RAFTADDRSRESPONSE._serialized_end=694
  _ASSIGNMENTRESPONSE._serialized_start=696
  _ASSIGNMENTRESPONSE._serialized_end=791
  _NEWLEADERRESPONSE._serialized_start=793
  _NEWLEADERRESPONSE._serialized_end=824
# @@protoc_insertion_point(module_scope)



import unittest
import logging
from unittest.mock import MagicMock, ANY
from .lan_queue import LANPrintQueue
from .sync_objects_test import TestReplDict

logging.basicConfig(level=logging.DEBUG)

class MockLockManager():
    def __init__(self, selfID=None):
        self.id = selfID
        self.locks = {}

    def tryAcquire(self, lockid, sync=False, timeout=0):
        if lockid in self.locks:
            return False
        self.locks[self.id] = [lockid]
        return True

    def release(self, lockid):
        for peer, locks in self.locks.items():
            locks.remove(lockid)

    def getPeerLocks(self):
        return self.locks

class LANQueueLocalTest():
    def setUp(self):
        self.addr = "localhost:6789"
        self.manifest = {"man": "ifest"}
        cb = MagicMock()
        self.q = LANPrintQueue("ns", self.addr, cb, logging.getLogger())
        self.q._init_base()
        self.q.q._syncobj = MagicMock()
        self.q.q.peers = TestReplDict(cb)
        self.q.q.jobs = TestReplDict(cb)
        self.q.q.locks = MockLockManager(self.addr)
    
    def tearDown(self):
        self.q.destroy()



class TestLanQueueInitExceptions(unittest.TestCase):
    def test_init_privileged_port(self):
        with self.assertRaises(ValueError):
            LANPrintQueue("ns", "locahost:80", None, logging.getLogger())

    def test_init_bad_addr(self):
        with self.assertRaises(ValueError):
            LANPrintQueue("ns", "hi", None, logging.getLogger())

class TestLanQueuePreStartup(unittest.TestCase):
    def setUp(self):
        self.q = LANPrintQueue("ns", "localhost:6789", MagicMock(), logging.getLogger())

    def tearDown(self):
        self.q.destroy()

    def test_startup_with_no_peers(self):
        self.q._init_base() # Don't call on_startup_complete as it actually does networking
        self.assertNotEqual(self.q.q, None)

    def test_startup_host_added(self):
        self.q._on_host_added("doesnothing") # Verifies no errors due to queue not initialized
        self.q._init_base()
        self.assertNotEqual(self.q.q, None)

class TestLanQueueOperations(LANQueueLocalTest, unittest.TestCase):
    def test_peer_added_after_startup(self):
        self.q._on_host_added("peer1")
        self.q.q._syncobj.addNodeToCluster.assert_called_with("peer1", callback=ANY)
        
    def test_peer_removed_after_startup(self):
        self.q._on_host_removed("peer1")
        self.q.q._syncobj.removeNodeFromCluster.assert_called_with("peer1", callback=ANY)

    def test_job_operations(self):
        self.q.q.setJob("hash", self.manifest)
        self.assertEqual(self.q.q.jobs["hash"], (self.addr, self.manifest))
        self.q.q.acquireJob("hash")
        self.assertEqual(self.q.q.locks.getPeerLocks()[self.addr][0], 'hash')
        self.q.q.releaseJob("hash")
        self.assertEqual(len(self.q.q.locks.getPeerLocks()[self.addr]), 0)
        self.q.q.removeJob("hash")
        self.assertEqual(len(self.q.q.jobs), 0)

    def testPeerGettersHideTimestamp(self):
        self.q.q.syncPeer(dict(a=1), "addr1")
        self.assertEqual(self.q.q.getPeers(), {"addr1": {"a": 1}})
        self.assertEqual(self.q.q.getPeer("addr1"), {"a": 1})

    def testJobGettersIncludePeer(self):
        self.q.q.setJob("job1", dict(a=1), addr="1.2.3.4")
        self.assertEqual(list(self.q.q.getJobs()), [("job1", ("1.2.3.4", {"a": 1}))])
        self.assertEqual(self.q.q.getJob("job1"), ("1.2.3.4", {"a": 1}))



from pysyncobj import SyncObj, SyncObjConf, FAIL_REASON, SyncObjException
from typing import Optional
from collections import defaultdict
from dataclasses import dataclass
import logging
import random
import time
from enum import Enum
from collections import defaultdict

from .discovery import P2PDiscovery
from .filesharing import pack_job
from .sync_objects import CPOrderedReplDict, CPReplLockManager

class ChangeType(Enum):
    PEER = "peer"
    JOB = "job"
    LOCK = "lock"
    QUEUE = "queue"

# This queue is shared with other printers on the local network which are configured with the same namespace.
# Actual scheduling and printing is done by the object owner.
# Job data should be considered opaque at this level - 
class LANPrintQueueBase():
    PEER_TIMEOUT = 60

    def __init__(self, ns, addr, update_cb, logger):
      self._logger = logger
      self._logger.debug("LANPrintQueueBase init")
      self.ns = ns
      self.acquire_timeout = 5
      self.addr = addr
      self.update_cb = update_cb 
      self._syncobj = None

    def _tagged_cb(self, changetype):
        def tagcb(prev, nxt):
            # Unwrap metadata tuples and return end value (user provided)
            if prev is not None and type(prev) is tuple:
                prev = prev[-1]
            if nxt is not None and type(nxt) is tuple:
                nxt = nxt[-1]
            self.update_cb(changetype, prev, nxt)
        return tagcb

    def connect(self, peers):
        # Peer dict is keyed by the peer addr. Value is tuple(last_update_ts, state)
        # where state is an opaque dict.
        self.peers = CPOrderedReplDict(self._tagged_cb(ChangeType.PEER))

        # Job dict is keyed by the ID of the job. Value is tuple(submitting_peer_addr, manifest)
        # where manifest is an opaque dict.
        self.jobs = CPOrderedReplDict(self._tagged_cb(ChangeType.JOB))

        self.locks = CPReplLockManager(selfID=self.addr, autoUnlockTime=60, cb=self._tagged_cb(ChangeType.LOCK))
        conf = SyncObjConf(
                onReady=self.on_ready, 
                dynamicMembershipChange=True,
          )
        self._syncobj = SyncObj(self.addr, peers, conf, consumers=[self.peers, self.jobs, self.locks])

    def is_ready(self):
        return self._syncobj.isReady()

    def on_ready(self):
        # Set ready state on all objects, enabling callbacks now
        # that they've been fast-forwarded
        self._logger.info("LANPrintQueueBase.on_ready")
        self.update_cb(ChangeType.QUEUE, False, True)

    # ==== Network methods ====

    def destroy(self):
        if self._syncobj is not None:
            self._syncobj.destroy()

    def addPeer(self, addr: str):
      self._logger.info(f"{self.ns}: Adding peer {addr}")
      self._syncobj.addNodeToCluster(addr, callback=self.queueMemberChangeCallback)
      self.syncPeer({}, addr)

    def removePeer(self, addr: str):
      self._logger.info(f"{self.ns}: Removing peer {addr}")
      self._syncobj.removeNodeFromCluster(addr, callback=self.queueMemberChangeCallback)
      self.peers.pop(addr, None)

    def queueMemberChangeCallback(self, result, error):
      if error != FAIL_REASON.SUCCESS:
        self._logger.error(f"membership change error: {result}, {error}")

    # ==== Mutation methods ====

    def syncPeer(self, state: dict, addr=None):
      if addr is None:
          addr = self.addr
      self.peers[addr] = (time.time(), state)
      for (addr, val) in self.peers.items():
          if val[0] < time.time() - self.PEER_TIMEOUT:
              self.peers.pop(addr, None)
    
    def getPeers(self):
      result = {}
      peerlocks = self.locks.getPeerLocks()
      for k, v in self.peers.items():
          result[k] = dict(**v[1]) # Exclude peer update timestamp
      return result

    def getPeer(self, peer):
        p = self.peers.get(peer)
        if p is not None:
            return dict(**p[-1])

    def hasJob(self, jid) -> bool:
        return (jid in self.jobs)

    def setJob(self, jid, manifest: dict, addr=None):
      # performed synchronously to prevent race conditions when quickly
      # writing, then reading job information
      if addr == None:
          addr = self.addr
      self.jobs.set(jid, (addr, manifest), sync=True, timeout=5.0)

    def getLocks(self):
        joblocks = {}
        for (peer, locks) in self.locks.getPeerLocks().items():
            for lock in locks:
                joblocks[lock] = peer
        return joblocks

    def getJobs(self):
        return self.jobs.ordered_items()

    def getJob(self, jid):
        return self.jobs.get(jid)

    def removeJob(self, jid: str):
        return self.jobs.pop(jid, None)

    def acquireJob(self, jid: str):
        try:
            return self.locks.tryAcquire(jid, sync=True, timeout=self.acquire_timeout)
        except SyncObjException: # timeout
            return False

    def releaseJob(self, jid: str):
        self.locks.release(jid)


# Wrap LANPrintQueueBase in a discovery class, allowing for dynamic membership based 
# on a namespace instead of using a list of specific peers.
class LANPrintQueue(P2PDiscovery):
  def __init__(self, ns, addr, update_cb, logger):
    super().__init__(ns, addr)

    (host, port) = addr.rsplit(':', 1)
    port = int(port)
    if port < 1024:
        raise ValueError(f"Queue {ns} must use a non-privileged port (want >1023, have {port})")

    self._logger = logger
    self.addr = addr
    self.ns = ns
    self.update_cb = update_cb
    self.q = None

  def connect(self):
    self._logger.info(f"Starting discovery for {self.ns} ({self.addr})")
    self.spin_async()

  def destroy(self):
    self._logger.info(f"Destroying discovery and SyncObj")
    super().destroy()
    self.q.destroy()

  def _on_host_added(self, host):
    if self.q is not None:
      self._logger.info(f"Host added: {host}")
      self.q.addPeer(host)

  def _on_host_removed(self, host):
    if self.q is not None:
      self._logger.info(f"Host removed: {host}")
      self.q.removePeer(host)

  def _init_base(self):
    self.q = LANPrintQueueBase(self.ns, self.addr, self.update_cb, self._logger)

  def _on_startup_complete(self, results):
    self._logger.info(f"Discover end: {results}; initializing queue")
    self._init_base()
    self.q.connect(results.keys())




import threading
import zmq
import inspect
import peerprint.server.proto.state_pb2 as spb
import peerprint.server.proto.peers_pb2 as ppb
import peerprint.server.proto.jobs_pb2 as jpb
from google.protobuf.message import Message
from google.protobuf.any_pb2 import Any

class MessageUnpackError(Exception):
    pass

class ZMQCallError(Exception):
    pass

class ZMQLogSink():
    def __init__(self, addr, logger):
        self._logger = logger
        self._context = zmq.Context()
        self._log_sock = self._context.socket(zmq.PULL)
        self._log_sock.bind(addr)
        self._log_thread = threading.Thread(target=self._stream_log, daemon=True)
        self._log_thread.start()
        self._logger.debug(f"ZMQLogSink thread started (bound to {addr})")
    
    def _stream_log(self):
        while self._log_sock is not None and not self._log_sock.closed:
            msg = self._log_sock.recv().decode('utf8').rstrip()
            self._logger.info(msg)

    def destroy(self):
        self._log_sock.destroy()
        self._log_sock = None


class ZMQClient():
    def __init__(self, req_addr, pull_addr, cb, logger):
        self._logger = logger
        self._cb = cb
        self._context = zmq.Context()
        self._sock = self._context.socket(zmq.REQ)
        self._pull = self._context.socket(zmq.PULL)
        self._sock.connect(req_addr) # connect to bound REP socket in golang code
        self._pull.bind(pull_addr) # bind for connecting PUSH socket in golang code
        self._logger.debug(f"ZMQClient connect to REQ {req_addr}")

        self._msgclss = []
        for k,p in [kp for m in (spb, jpb, ppb) for kp in m.__dict__.items()]:
            if inspect.isclass(p) and issubclass(p, Message):
                self._msgclss.append(p)
        self._logger.debug(f"Loaded {len(self._msgclss)} message classes")

        self._pull_thread = threading.Thread(target=self._listen, daemon=True)
        self._pull_thread.start()
        self._logger.debug(f"ZMQClient listener thread started (bound to {pull_addr})")
    
    def destroy(self):
        self._sock.destroy()
        self._pull.destroy()

    def _unpack(self, data):
        apb = Any()
        apb.ParseFromString(data)
        for p in self._msgclss:
            if apb.Is(p.DESCRIPTOR):
                m = p()
                apb.Unpack(m)
                if p == spb.Error:
                    raise ZMQCallError(m.status)
                return m
        raise MessageUnpackError(f"Could not unpack message: {apb}")

    def _listen(self):
        while self._pull is not None and not self._pull.closed:
            data = self._pull.recv()
            try:
                self._cb(self._unpack(data))
            except MessageUnpackError as e:
                self._cb(e)

    def call(self, p):
        amsg = Any()
        amsg.Pack(p)
        self._sock.send(amsg.SerializeToString())
        data = self._sock.recv()
        return self._unpack(data)



import unittest
import logging
from unittest.mock import MagicMock
from .wan_queue import PeerPrintQueue
from .proc import ServerProcessOpts
import peerprint.server.proto.peers_pb2 as ppb
import peerprint.server.proto.state_pb2 as spb
import peerprint.server.proto.jobs_pb2 as jpb

logging.basicConfig(level=logging.DEBUG)

class ObjectCodec:
    @classmethod
    def encode(self, data):
        return b"", "object"

    @classmethod
    def decode(self, data, protocol):
        assert protocol=="object"
        return dict()

class TestPeerPrintQueue(unittest.TestCase):
    def setUp(self):
        self.cb = MagicMock()
        self.q = PeerPrintQueue(
                opts=ServerProcessOpts(),
                peer_id="foo",
                codec=ObjectCodec,
                binary_path="testbinary",
                update_cb=self.cb,
                logger=logging.getLogger(),
                keydir=None,
            )
        self.q._zmqclient = MagicMock()

    def tearDown(self):
        self.q.destroy()

    def testSyncPeer(self):
        self.q.syncPeer(dict(asdf="ghjk"), addr="abc:123")
        self.q._zmqclient.call.assert_called_with(ppb.PeerStatus())

    def testGetPeersNone(self):
        got = self.q.getPeers()
        self.assertEqual(got.sample, [])

    def testGetPeersWithPeerList(self):
        want = ppb.PeersSummary(peer_estimate=1, variance=float('inf'), sample=[ppb.PeerStatus(id="foo")])
        self.q._update(want)
        self.assertEqual(self.q.getPeers(), want)

    def testJobPresenceAndGetters(self):
        s=spb.State(jobs=dict(foo=jpb.Job(id="foo", data=b"", protocol="object", owner='testpeer')))
        self.q._update(s)
        self.assertEqual(self.q.hasJob("foo"), True)
        self.assertEqual(self.q.hasJob("bar"), False)
        self.assertEqual(self.q.getJobs(), {'foo': {'peer_': 'testpeer'}})

    def testSetJob(self):
        self.q.setJob("foo", dict(man="ifest"), addr="testaddr")
        req = self.q._zmqclient.call.call_args[0][0]
        self.assertEqual(req.job.id, "foo")

    def testRemoveJob(self):
        self.q.removeJob("foo")
        self.q._zmqclient.call.assert_called_with(jpb.DeleteJobRequest(id="foo"))

    def testAcquireJob(self):
        self.q.acquireJob("foo")
        self.q._zmqclient.call.assert_called_with(jpb.AcquireJobRequest(id="foo"))

    def testReleaseJob(self):
        self.q.releaseJob("foo")
        self.q._zmqclient.call.assert_called_with(jpb.ReleaseJobRequest(id="foo"))
